// (C) Copyright 2002-2007 by Autodesk, Inc. 
//
// Permission to use, copy, modify, and distribute this software in
// object code form for any purpose and without fee is hereby granted, 
// provided that the above copyright notice appears in all copies and 
// that both that copyright notice and the limited warranty and
// restricted rights notice below appear in all supporting 
// documentation.
//
// AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS. 
// AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
// MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC. 
// DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
// UNINTERRUPTED OR ERROR FREE.
//
// Use, duplication, or disclosure by the U.S. Government is subject to 
// restrictions set forth in FAR 52.227-19 (Commercial Computer
// Software - Restricted Rights) and DFAR 252.227-7013(c)(1)(ii)
// (Rights in Technical Data and Computer Software), as applicable.
//

using System;
using Autodesk.AutoCAD.Runtime;
using Autodesk.AutoCAD.DatabaseServices;
using Autodesk.AutoCAD.ApplicationServices;
using Autodesk.AutoCAD.EditorInput;
using Autodesk.AutoCAD.BoundaryRepresentation;
using Autodesk.AutoCAD.Geometry;
using Autodesk.AutoCAD.GraphicsInterface;
using System.Runtime.InteropServices;

[assembly: CommandClass(typeof(BRepDemo.ADSCommands))]

namespace BRepDemo
{
    public class LibWrap
    {
        
        // Declaration to P/Invoke the "body" method on the AcDbSurface object
        [DllImport("acdb17.dll",
      EntryPoint = "?body@AcDbSurface@@UBEPAXXZ",
       CallingConvention = CallingConvention.ThisCall)]
        public static extern IntPtr GetSurfBody(IntPtr ths);

        // Declaration to P/Invoke the "setBody" method on the AcDbBody object
        [DllImport("acdb17.dll",
      EntryPoint = "?setBody@AcDbBody@@UAE?AW4ErrorStatus@Acad@@PBX@Z",
       CallingConvention = CallingConvention.ThisCall)]
        public static extern int CreateSurfBdy(IntPtr ths,IntPtr bdyPtr);

        
    }

    /// <summary>
    /// Summary description for ADSCommands.
    /// </summary>
    public class ADSCommands
    {
        public ADSCommands()
        {
            //
            // TODO: Add constructor logic here
            //
        }

        // Define Command "AsdkCmd1"
        [CommandMethod("Traversal")]
        static public void Traversal() // This method can have any name
        {
            // Code to demonstrate traversal of a solid

            // Obtain the necessary objects, Database, editor and transaction
            Database db = HostApplicationServices.WorkingDatabase;
            Editor ed = Application.DocumentManager.MdiActiveDocument.Editor;

            Transaction tr = db.TransactionManager.StartTransaction();

            try
            {
                // Prompt for selection of a solid to be traversed
                PromptEntityOptions prEntOpt = new PromptEntityOptions("Select a 3D solid: ");
                prEntOpt.SetRejectMessage("\nPlease select only a 3D Solid");
                prEntOpt.AddAllowedClass(typeof(Solid3d), true);

                PromptEntityResult prEntRes = ed.GetEntity(prEntOpt);
                ObjectId[] objIds = {prEntRes.ObjectId};
                
                //FullSubentityPath fullSubPath = new FullSubentityPath(objIds, SubentityId.Null);
                //Brep brp = new Brep(fullSubPath);

                Solid3d sol = (Solid3d)tr.GetObject(prEntRes.ObjectId, OpenMode.ForRead);

                // Build the BRep topology object that will be traversed. 
                Brep brp = new Brep(sol);
                int cmplxCount=0, shllCount=0, fceCount=0, lpCount = 0, edgCount = 0;

                // Get all the Complexes which are primary BRep elements and represent a
                // conceptual topological entity of connected shell boundaries.
                foreach(Complex cmplx in brp.Complexes)
                {
                    ed.WriteMessage("\n\nComplex Count: {0}", ++cmplxCount);

                    // Get all the shells within a complex. Shells are secondary BRep entities
                    // that corresponds to a collection of neighboring surfaces on a solid
                    foreach(Shell shll in cmplx.Shells)
                    {
                        ed.WriteMessage("\nShell Count: {0}", ++shllCount);
                        ed.WriteMessage("\nShell Type: {0}", shll.ShellType.ToString());

                        // Get all the faces in a shell. Faces are primary BRep topological entities that directly
                        // correspond to face subentities on AutoCAD entities like solid, region and body
                        foreach (Autodesk.AutoCAD.BoundaryRepresentation.Face fce in shll.Faces)
                        {
                            ed.WriteMessage("\nFace Count: {0}", ++fceCount);

                            // Get all the boundary loops within a face (Secondary BRep entities and no corresponding
                            // AutoCAD entities/subentities).
                            foreach(BoundaryLoop lp in fce.Loops)
                            {
                                ed.WriteMessage("\nLoop Count: {0}", ++lpCount);
                                ed.WriteMessage("\nLoop Type: {0}", lp.LoopType.ToString());

                                // Get all the Edges in a loop (Edges are primary BRep entities and correspond to
                                // Geometric entities). Output the 
                                foreach(Edge edg in lp.Edges)
                                {
                                    ed.WriteMessage("\nEdge Count: {0}", ++edgCount);
                                    ed.WriteMessage("\nEdge Vertex 1: {0}", edg.Vertex1.Point.ToString());
                                    ed.WriteMessage("\nEdge Vertex 2: {0}\n", edg.Vertex2.Point.ToString());
                                }
                            }
                        }
                    }
                }

                tr.Commit();
            }
            catch(System.Exception ex)
            {
                ed.WriteMessage("\nException in traversal: {0}", ex.Message);
            }
            finally
            {
                tr.Dispose();
            }

        }

        [CommandMethod("containPt")]
        static public void containPt() // This method can have any name
        {
            //Code to determine the containment of a point within a solid

            // Obtain the necessary objects, Database, editor and transaction
            Database db = HostApplicationServices.WorkingDatabase;
            Editor ed = Application.DocumentManager.MdiActiveDocument.Editor;

            Transaction tr = db.TransactionManager.StartTransaction();

            try
            {
                // Prompt for a solid
                PromptEntityOptions prEntOpt = new PromptEntityOptions("Select a 3D solid: ");
                prEntOpt.SetRejectMessage("\nPlease select only a 3D Solid");
                prEntOpt.AddAllowedClass(typeof(Solid3d), true);

                PromptEntityResult prEntRes = ed.GetEntity(prEntOpt);
                ObjectId[] objIds = { prEntRes.ObjectId };

                //FullSubentityPath fullSubPath = new FullSubentityPath(objIds, SubentityId.Null);
                //Brep brp = new Brep(fullSubPath);
                
                // Create a BRep topology object from the selected Solid object
                Solid3d sol = (Solid3d)tr.GetObject(prEntRes.ObjectId, OpenMode.ForRead);
                Brep brp = new Brep(sol); // The Acad entity should be open as long as we intended
                                          // to use the BRep topology object

                // Select or input the point used to test containment
                PromptPointOptions prPtOpts = new PromptPointOptions("\nPlease input a point: ");
                prPtOpts.AllowArbitraryInput = false;
                PromptPointResult prPtRes = ed.GetPoint(prPtOpts);

                // Test for containment of the point in the selected solid
                PointContainment ptCont = new PointContainment();
                BrepEntity brepEnt = brp.GetPointContainment(prPtRes.Value, out ptCont);
                ed.WriteMessage("\nThe result of the point containment test is: {0}", ptCont);

                tr.Commit();

            }
            catch(System.Exception ex)
            {
                ed.WriteMessage("\nException in Point Containment code: {0}",ex.Message);
            }
            finally
            {
                tr.Dispose();
            }

        }

        [CommandMethod("containLn")]
        static public void containLn() // This method can have any name
        {
            //Code to determine the containment of a line within a solid

            // Obtain the necessary objects, Database, editor and transaction
            Database db = HostApplicationServices.WorkingDatabase;
            Editor ed = Application.DocumentManager.MdiActiveDocument.Editor;

            Transaction tr = db.TransactionManager.StartTransaction();

            try
            {
                // Prompt for a solid
                PromptEntityOptions prEntOpt = new PromptEntityOptions("Select a 3D solid: ");
                prEntOpt.SetRejectMessage("\nPlease select only a 3D Solid");
                prEntOpt.AddAllowedClass(typeof(Solid3d), true);

                PromptEntityResult prEntRes = ed.GetEntity(prEntOpt);
                ObjectId[] objIds = { prEntRes.ObjectId };

                //FullSubentityPath fullSubPath = new FullSubentityPath(objIds, SubentityId.Null);
                //Brep brp = new Brep(fullSubPath);

                // Create a BRep topology object based on the selected solid
                Solid3d sol = (Solid3d)tr.GetObject(prEntRes.ObjectId, OpenMode.ForRead);
                Brep brp = new Brep(sol); // The Acad entity should be open as long as we intended
                // to use the BRep topology object

                // Acquire points to build the line that we use to test containment.
                PromptPointOptions prPtOpts = new PromptPointOptions("\nPlease input start point: ");
                prPtOpts.AllowArbitraryInput = false;
                PromptPointResult prPtRes = ed.GetPoint(prPtOpts);

                prPtOpts.Message = "\nPlease input start point: ";
                PromptPointResult prPtRes2 = ed.GetPoint(prPtOpts);

                LinearEntity3d ln = new Line3d(prPtRes.Value,prPtRes2.Value); // This could be any linear entity
                                                                              // including Line3d, Ray3d and LineSegment3d

                // Test for line containment within the selected solid
                Hit[] hits = brp.GetLineContainment(ln,10);
                
                // Output points at which the line hits the solid.
                foreach(Hit hit in hits)
                {
                    ed.WriteMessage("\nPoint Hit: {0}", hit.Point);
                }

                tr.Commit();

            }
            catch (System.Exception ex)
            {
                ed.WriteMessage("\nException in Point Containment code: {0}", ex.Message);
            }
            finally
            {
                tr.Dispose();
            }

        }

        [CommandMethod("meshEntity")]
        static public void meshEntity() // This method can have any name
        {
            //Code to mesh a selected solid

            // Obtain the necessary objects, Database, editor and transaction
            Database db = HostApplicationServices.WorkingDatabase;
            Editor ed = Application.DocumentManager.MdiActiveDocument.Editor;

            Transaction tr = db.TransactionManager.StartTransaction();

            try
            {

                // Prompt for a solid
                PromptEntityOptions prEntOpt = new PromptEntityOptions("Select a 3D solid: ");
                prEntOpt.SetRejectMessage("\nPlease select only a 3D Solid");
                prEntOpt.AddAllowedClass(typeof(Solid3d), true);

                PromptEntityResult prEntRes = ed.GetEntity(prEntOpt);
                Solid3d sol = (Solid3d)tr.GetObject(prEntRes.ObjectId, OpenMode.ForRead);
                Brep brp = new Brep(sol); // The Acad entity should be open as long as we intended

                // Set the Mesh2dControl settings
                Mesh2dControl msh2dCtrl = new Mesh2dControl();
                msh2dCtrl.MinSubdivisionsInU = 100;
                msh2dCtrl.MinSubdivisionsInV = 100;
                msh2dCtrl.MaxSubdivisions = 100000;
                msh2dCtrl.MaxNodeSpacing = 0.5;
                msh2dCtrl.AngleTolerance = 0.1;
                msh2dCtrl.DistanceTolerance = 0.1;
                msh2dCtrl.MaxAspectRatio = 2;
                msh2dCtrl.ElementShape = Element2dShape.Default;

                // Create a mesh filter that maps the topology object to the mesh control settings
                // used to generate the mesh
                Mesh2dFilter msh2dFilt = new Mesh2dFilter();
                msh2dFilt.Insert(brp, msh2dCtrl);

                // Generate the mesh
                Mesh2d msh2d = new Mesh2d(msh2dFilt);

                // Display the mesh using polylines using the transient graphics API
                foreach(Element2d ele2d in msh2d.Element2ds)
                {
                    Point3dCollection pts = new Point3dCollection();
                    foreach(Node nd in ele2d.Nodes)
                    {
                        pts.Add(nd.Point);
                    }

                    Polyline3d pl3d = new Polyline3d(Poly3dType.SimplePoly, pts, true);
                    IntegerCollection intCol = new IntegerCollection();
                    TransientManager.CurrentTransientManager.AddTransient(pl3d, TransientDrawingMode.DirectShortTerm, 128, intCol);
                    pl3d.Dispose();
                }
                
                tr.Commit();

            }
            catch(System.Exception ex)
            {
                ed.WriteMessage("Exception in Meshing Code: {0}", ex.Message);
            }
            finally
            {
                tr.Dispose();
            }

        }

        [CommandMethod("SurfaceIntero")]
        static public void SurfaceIntero() // This method can have any name
        {

            // Method to interrogate the AutoCAD Surface object using the B-Rep API

            // Obtain the necessary objects, Database, editor and transaction
            Database db = HostApplicationServices.WorkingDatabase;
            Editor ed = Application.DocumentManager.MdiActiveDocument.Editor;

            Transaction tr = db.TransactionManager.StartTransaction();

            try
            {
                // Prompt to select a surface
                PromptEntityOptions prEntOpt = new PromptEntityOptions("Select a Surface: ");
                prEntOpt.SetRejectMessage("\nPlease select only a Surface");
                prEntOpt.AddAllowedClass(typeof(Autodesk.AutoCAD.DatabaseServices.Surface), true);

                PromptEntityResult prEntRes = ed.GetEntity(prEntOpt);
                Autodesk.AutoCAD.DatabaseServices.Surface surf = (Autodesk.AutoCAD.DatabaseServices.Surface)tr.GetObject(prEntRes.ObjectId, OpenMode.ForRead);
                
                // Use P/Invoke to obtain the unmanaged body pointer for the surface object. We need to do this
                // because unfortunately, the BRep .NET API does not directly reference the body object within the surface
                IntPtr bdyPtr = LibWrap.GetSurfBody(surf.UnmanagedObject);
                //(Body)Body.Create(bdyPtr, false);

                // Create a new Body object
                Body bdy = new Body();

                // Use P/Invoke to set the newly created body object with the data from the 
                // Surface body object
                LibWrap.CreateSurfBdy(bdy.UnmanagedObject, bdyPtr);
                

                // Now we can create the BRep topology object
                Brep brp = new Brep(bdy);
                
                //ObjectId[] objIds = { prEntRes.ObjectId };
                //FullSubentityPath fullSubPath = new FullSubentityPath(objIds, SubentityId.Null);
                //BrepEntity brp = new Autodesk.AutoCAD.BoundaryRepresentation.Face(fullSubPath);
                
                
                // Find the area of the faces on the Surface
                foreach(Autodesk.AutoCAD.BoundaryRepresentation.Face fc in brp.Faces)
                {
                    ed.WriteMessage("\nThe surface area: {0}",fc.GetSurfaceArea());

                    //fc.Dispose();
                }
               

                //brp.Dispose();
                tr.Commit();
                bdy.Dispose();

            }
            catch(System.Exception ex)
            {
                ed.WriteMessage("Exception in SurfaceTrav: {0}",ex.Message);

            }
            finally
            {
                tr.Dispose();
            }

        }

        [CommandMethod("MiscCode")]
        static public void MiscCode() // This method can have any name
        {
            Database db = HostApplicationServices.WorkingDatabase;
            Editor ed = Application.DocumentManager.MdiActiveDocument.Editor;
            Transaction tr = db.TransactionManager.StartTransaction();

            try
            {
                PromptSelectionOptions prSelOpts = new PromptSelectionOptions();

                prSelOpts.AllowSubSelections = true;
                prSelOpts.SingleOnly = true;

                //Selection Set Filter for Solid3d 
               TypedValue[] values = {new TypedValue((int)DxfCode.Start, "3DSOLID")};

                SelectionFilter selFil = new SelectionFilter(values);

                PromptSelectionResult prSelRes = ed.GetSelection(prSelOpts, selFil);

                SelectionSet selSet = prSelRes.Value;

                SelectedObject selObj = selSet[0];
                //selObj.GraphicsSystemMarker;
                


            }
            catch(System.Exception ex)
            {
                ed.WriteMessage("\nMiscellaneous Exception: {0}", ex.Message);
            }
            finally
            {
                tr.Dispose();
            }

        }

    }
}